# Copyright (c) 2016-2024 The Regents of the University of Michigan
# Part of GSD, released under the BSD 2-Clause License.

"""Define the current version of the gsd package.

Attributes:
    version (str): GSD software version number. This is the version number
                   of the software package as a whole,
                   not the file layer version it reads/writes.
"""

version = '3.4.2'

__all__ = [
    'version',
]
