# Copyright (c) 2016-2024 The Regents of the University of Michigan
# Part of GSD, released under the BSD 2-Clause License.

"""Global pytest options."""

pytest_plugins = ('gsd.pytest_plugin_validate',)
